module GameOfGame::Climate

  # Econnomy and productivity:
  # Betting is producing waste.
  # Producing waste multiply the bet.
  # There is a basic income every turn added to the bet, which do not pollute.
  #
  # Climate change impact on the economy:
  # Wastes accumulate with very little decay.
  # Waste reduce production every futur round for everyone.
  #
  # Economic status is inequal:
  # Some players will randomly have assigned a bonus to production.
  #
  # Climate change is inequal:
  # Some players will randomly have assigned a bonus or malus to the impact of
  # waste accumulationg on its production.
  class Version01 < RulesSet
    @round : Round? = nil

    @players_production_coef = Hash(Ai, Float64).new
    @players_climate_impact_coef = Hash(Ai, Float64).new

    getter rounds_amount
    getter rounds_amount : Int32
    getter players : Array(Ai)
    getter base_prod : Float64
    getter waste_coef : Float64
    getter total_wastes : Float64
    getter economic_inequality_factor : Float64
    getter climate_inequality_factor : Float64

    def initialize(
         @rounds_amount,                      # unused
         @players,                            # initial player list
         @base_prod                  = 1.0,   # base bonus to the bet, without waste
         @waste_coef                 = 0.001, # a waste inflict a linear malus to the economy
         @total_wastes               = 0.0,   # initial accumulated
         @economic_inequality_factor = 10.0,  # actors disparity, multiply the total prod
         @climate_inequality_factor  = 3.0,   # actors disparity, multiply the waste coef
         name : String? = nil,
       )

      step_production_coef = @economic_inequality_factor / @players.size
      @players.shuffle.each.with_index do |player, index|
        @players_production_coef[player] = (index + 1.0) * step_production_coef
      end

      step_climate_impact_coef = @climate_inequality_factor / @players.size
      @players.shuffle.each.with_index do |player, index|
        @players_climate_impact_coef[player] = (index + 1.0) * step_climate_impact_coef
      end

      super(name)
    end

    def examine(round : Round, player : Ai) : Float64
      if @round != round
        new_round(round)
      end

      examine_current_round(player, round)
    end


    def new_round(round : Round)
      @round = round
      if (previous = round.previous)
        added_wastes = previous.bets.values.sum
        @total_wastes += added_wastes
      end
    end

    def examine_current_round(player : Ai, round : Round) : Float64
      total_prod = (@base_prod + round.bets[player]) * @players_production_coef[player]
      total_malus = @total_wastes * @waste_coef * @players_climate_impact_coef[player]
      total_prod - total_malus
    end
  end

  class FullNature < Ai
    def bet(rounds : GameHistory) : Float64
      0.0
    end
  end

  class FullProd < Ai
    def bet(rounds : GameHistory) : Float64
      1.0
    end
  end

  class LinearReducer < Ai
    def initialize(@zero_net : Float64)
      super()
    end

    def bet(rounds : GameHistory) : Float64
      return 0.0 if rounds.history.size > @zero_net
      1.0 - (rounds.history.size / @zero_net)
    end
  end
end
